from time import gmtime, time, strptime, strftime, mktime, timezone

import logging
logger = logging.getLogger(__name__)


def epoch_to_iso8601(timestamp):
    try: mlsec = str(timestamp).split('.')[1][:3]
    except: mlsec = "000"
    mlsec = mlsec + "0"*(3-len(mlsec))
    iso_str = strftime("%Y-%m-%dT%H:%M:%S.{}Z".format(mlsec), gmtime(timestamp))
    return iso_str

def iso8601_to_epoch(iso_str):
    str_time, mlsec = iso_str.split('.')
    mlsec = float(mlsec[:-1])/1000
    time_tuple = strptime(str_time+"UTC", "%Y-%m-%dT%H:%M:%S%Z")
    return mktime(time_tuple) - timezone + mlsec

def epoch_to_pathfriendly(timestamp=None):
    if timestamp is None: timestamp = time()
    return strftime("%Y%m%d%H%M%S", gmtime(timestamp))





#submodules
from easilyb.net import requestqueue
#imports
from subprocess import Popen as popen
from random import choice as _rchoice
from time import time as _time
from logging import getLogger
logger = getLogger(__name__)

_ANDROID_VERSIONS = ['3.2', '3.2.1', '3.2.2', '3.2.3', '3.2.4', '3.2.5', '3.2.6', '4.0', '4.0.1', '4.0.2', '4.0.3', '4.0.4', '4.1', '4.1.1', '4.1.2', '4.2', '4.2.1', '4.2.2', '4.3', '4.3.1', '4.4', '4.4.1', '4.4.2', '4.4.3', '4.4.4', '5.0', '5.0.1', '5.0.2', '5.1', '5.1.1', '6.0', '6.0.1']
_FIREFOX_VERSIONS = None
_FIREFOX_M0 = 560.0
_FIREFOX_ANDROID_AGENT = 'Mozilla/5.0 (Android %s; Mobile; rv:%s) Gecko/%s Firefox/%s'

def wget(url, filepath=None, referer=None, useragent=None, basedir=None, echo=False):
    if useragent is None:
        useragent = get_useragent()
    cmd = ['wget', url ]
    if filepath:
        cmd += ['-O', filepath]
    elif basedir:
        cmd += ['-P', basedir]
    cmd += ['-U', useragent]
    if referer:
        cmd += ['--referer', referer]
    if echo:
        cmd = ['echo'] + cmd
    logger.debug("wget command: %s", cmd)
    try:
        return popen(cmd).wait()
    except:
        logger.warning('error running wget.', exc_info=True)
        return 999
def _generate_firefox_versions():
    global _FIREFOX_VERSIONS
    t1 = _time()
    m1 = t1/3600/24/30
    _max = int(44 +(m1 - _FIREFOX_M0)*2.0/3.0)
    _min = max(_max-16, 10)
    _FIREFOX_VERSIONS = [str(i)+'.0' for i in range(_min, _max)]

def get_useragent():
    if _FIREFOX_VERSIONS is None:
        _generate_firefox_versions()
    fv = _rchoice(_FIREFOX_VERSIONS)
    av = _rchoice(_ANDROID_VERSIONS)
    return _FIREFOX_ANDROID_AGENT%(av, fv, fv, fv)


import logging
from logging.config import dictConfig

DATEFMT = '%Y-%m-%d %H:%M:%S'
FORMAT = "%(asctime)s %(levelname)-8s:%(name)s:%(message)s"
FORMAT_NO_TIME = "%(levelname)-8s:%(name)s:%(message)s"


COLORS = {
    'DEBUG': '\033[1m\033[34m', # bold + blue
    'INFO': '\033[1m', # bold
    'WARNING': '\033[1m\033[33m', # yellow + bold
    'WARN': '\033[1m\033[33m', # yellow + bold
    'ERROR': '\033[31m',  # red
    'EXCEPTION': '\033[31m',  # red
    'CRITICAL': '\033[4m\033[1m\033[31m' ,  # red + underline + bold
    'FATAL': '\033[4m\033[1m\033[31m' ,  # red + underline + bold
    'RESET': "\033[0m" # reset colors
}


def configure_logging(console=True, level='DEBUG', modules=None, other_modules_level='WARNING', filepath=None,
                      max_bytes=10485760, backup_count=5, console_format=None, file_format=None,
                      console_datefmt=None, file_datefmt=None, console_print_time=False,
                      console_colored=True, console_level=None, file_level=None, file_mode='a', other_level_rules=None):
    handlers = {}
    formatters = {}
    loggers = {}
    # Console handler
    if console:
        console_handler = {'class': 'logging.StreamHandler', 'level': level, 'formatter': 'console_formatter'}
        if console_level is not None:
            console_handler['level'] = console_level
        handlers['console_handler'] = console_handler

        # console_formatter
        console_formatter = {}
        if console_colored:
            console_formatter['()'] =  ColoredFormatter
        if console_format is None:
            if console_print_time:
                console_formatter['format'] = FORMAT
            else:
                console_formatter['format'] = FORMAT_NO_TIME
        else:
            console_formatter['format'] = console_format
        console_formatter['datefmt'] = console_datefmt if console_datefmt is not None else DATEFMT
        formatters['console_formatter'] = console_formatter

    # File handler
    if filepath is not None:
        file_handler = {
            'class': 'logging.handlers.RotatingFileHandler',
            'level': level,
            'formatter': 'file_formatter',
            'filename': filepath,
            'mode': file_mode,
            'maxBytes': max_bytes,
            'backupCount': backup_count,
        }
        if file_level is not None:
            file_handler['level'] = file_level
        handlers['file_handler'] = file_handler

        # file formatter
        file_formatter = {}
        file_formatter['format'] = file_format if file_format is not None else FORMAT
        file_formatter['datefmt'] = file_datefmt if file_datefmt is not None else DATEFMT
        formatters['file_formatter'] = file_formatter

    # root logger
    loggers[''] = {'level':other_modules_level, 'handlers':handlers.keys()}

    # main logger
    loggers['__main__'] = {'level':level}

    # modules loggers
    if modules is None:
        modules = list()
    for m in modules:
        loggers[m] = {'level':level}

    # other modules rules:
    if other_level_rules is not None:
        for k,v in other_level_rules.items():
            loggers[k] = v

    # logging dict config
    config = {
        'version': 1,
        'handlers': handlers,
        'formatters': formatters,
        'loggers': loggers
    }

    dictConfig(config)


def _wrap_level_color(txt, level):
    return COLORS[level] + txt + COLORS['RESET']


class ColoredFormatter(logging.Formatter):
    def __init__(self, format=FORMAT, datefmt=DATEFMT):
        super(ColoredFormatter, self).__init__(format, datefmt=datefmt)

    def format(self, record):
        msg, levelname = record.msg, record.levelname
        if record.levelname in COLORS:
            record.msg = _wrap_level_color(msg, levelname)
            record.levelname = _wrap_level_color(levelname, levelname)
        s = logging.Formatter.format(self, record)
        record.msg, record.levelname = msg, levelname
        return s


def test_logging(loggername=''):
    logger = logging.getLogger(loggername)
    logger.debug('Jackdaws love my big sphinx of quartz.')
    logger.info('Jackdaws love my big sphinx of quartz.')
    logger.warning('Jackdaws love my big sphinx of quartz.')
    logger.error('Jackdaws love my big sphinx of quartz.')
    logger.critical('Jackdaws love my big sphinx of quartz.')



import os
import socket
import threading
import logging
import easilyb.json_serialize as json

logger = logging.getLogger(__name__)
BUFFER_SIZE = 4096
MAX_OBJECT_SIZE = 1024 * 1024 * 1024  # 1GB
DEFAULT_MAX_TRIES = 5


class JsonSocket(object):
    def __init__(self, sock, max_tries=DEFAULT_MAX_TRIES):
        self.sock = sock
        self.max_tries = max_tries

    def _send(self, data):
        tries = 0
        while tries < self.max_tries:
            # Print('sending... ', end='')
            self.sock.sendall(data.encode())
            # Print('sent: ', data)
            # Print('receiving... ', end='')
            ret = self.sock.recv(1).decode()
            # Print('received : ' + ret)
            if ret == 'k':
                return
            elif ret == 'r':
                pass
            else:
                raise IOError('Excpected k or r from peer')
            tries += 1

    def _send_fragmented(self, obj_ser, size):
        tries = 0
        fragement_size = BUFFER_SIZE - 2
        while tries < self.max_tries:
            self._send('s' + str(size) + '\\')  # send size s555\
            # print 's' + str(size) + '\\'
            i = 0
            while i < size:
                self._send('c' + obj_ser[i:i + fragement_size] + '\\')
                # print 'c' + str(len(obj_ser[i:i + fragement_size])) + '\\'
                i += fragement_size
            # Print('sending... ', end='')
            self.sock.sendall(b'F\\')
            # Print('sent F\\')
            # print 'F\\'
            # Print('receiving... ', end='')
            ret = self.sock.recv(1).decode()
            # Print('received : ' + ret)
            if ret == 'K':
                return
            elif ret == 'R':
                pass
            else:
                raise IOError('Excpected K or R from peer')
            tries += 1

    def send(self, obj):
        # logger.debug("sending obj: %s", obj)
        obj_ser = json.dumps(obj)
        size = len(obj_ser)
        tries = 0
        if size > MAX_OBJECT_SIZE:
            raise ValueError("Can't send all this data!")
        try:
            if size < BUFFER_SIZE - 2:
                self._send('1' + obj_ser + '\\')
            else:
                self._send_fragmented(obj_ser, size)
        except Exception:
            # if tries > self.max_tries or not self.reconnect():
            if tries > self.max_tries:  #or not self.reconnect():
                raise
            else:
                tries += 1

    def receive(self):
        state = 0
        data = ''
        size = 0
        obj_ser = ''
        obj = None
        state2_tries = 0
        state8_tries = 0
        state12_tries = 0
        while True:
            if state == 0:
                # Print('receiving... ', end='')
                data = self.sock.recv(BUFFER_SIZE).decode()
                # Print('received : ' + data)
                if data.endswith('\\'):
                    state = 1
                else:
                    state = 2
            elif state == 1:
                if data.startswith('1'):
                    state = 3
                elif data.startswith('s'):
                    state = 4
                else:
                    state = 2
            elif state == 2:
                if state2_tries < self.max_tries:
                    state2_tries += 1
                    # Print('sending... ', end='')
                    self.sock.sendall(b'r')
                    # Print('sent k')

                    state = 0
                else:
                    raise IOError("Data malformated!")
            elif state == 3:
                obj_ser = data[1:-1]
                try:
                    obj = json.loads(obj_ser)
                    state = 5
                except:
                    state = 2
            elif state == 4:
                size = int(data[1:-1])
                obj_ser = ''
                # print 'received: s%d//' % (size)
                # Print('sending... ', end='')
                self.sock.sendall(b'k')
                # Print('sent k')
                state = 6
            elif state == 5:
                # Print('sending... ', end='')
                self.sock.sendall(b'k')
                # Print('sent k')
                # logger.debug('received obj: %s', obj)
                return obj
            elif state == 6:
                # Print('receiving... ', end='')
                data = self.sock.recv(BUFFER_SIZE).decode()
                # Print('received : ' + data)
                if data.endswith('\\'):
                    state = 7
                else:
                    state = 8
            elif state == 7:
                if data.startswith('c'):
                    # print 'received: c + %d'%(len(data)-2)
                    state = 9
                elif data.startswith('F'):
                    # print 'received: F'
                    state = 10
                else:
                    state = 8
            elif state == 8:
                if state8_tries < self.max_tries:
                    state8_tries += 1
                    # Print('sending... ', end='')
                    self.sock.sendall(b'r')
                    # Print('sent r')
                    state = 6
                else:
                    raise IOError("Data malformated!")
            elif state == 9:
                obj_ser += data[1:-1]
                # Print('sending... ', end='')
                self.sock.sendall(b'k')
                # Print('sent k')
                state = 6
            elif state == 10:
                if size == len(obj_ser):
                    try:
                        obj = json.loads(obj_ser)
                        state = 11
                    except:
                        state = 12
                else:
                    state = 12
            elif state == 11:
                # Print('sending... ', end='')
                self.sock.sendall(b'K')
                # Print('sent K')
                # logger.debug('received obj: %s', obj)
                return obj
            elif state == 12:
                if state12_tries < self.max_tries:
                    state12_tries += 1
                    # Print('sending... ', end='')
                    self.sock.sendall(b'R')
                    # Print('sent R')
                    state = 0
                else:
                    raise IOError("Data malformated!")

    def reconnect(self):
        # logger.debug("reconnecting...")
        return False

    def close(self):
        try:
            self.sock.close()
        except:
            pass
        finally:
            self.sock = None


class JsonServerSocket(JsonSocket):
    def __init__(self, server_address, callback):
        self.server_address = server_address
        self.callback = callback
        super(JsonServerSocket, self).__init__(None)  # does not work
        # super(JsonServerSocket, self).__init__(socket.socket(socket.AF_UNIX, socket.SOCK_STREAM))  # does not work
        # super call:
        # self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        # self.max_tries = DEFAULT_MAX_TRIES
        # end super call

    def bind(self):
        try:
            os.unlink(self.server_address)
        except OSError:
            if os.path.exists(self.server_address):
                raise
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sock.bind(self.server_address)
        logger.info("binded to %s" % self.server_address)
        self.sock.listen(1)

    def accept(self):
        connection, client_address = self.sock.accept()
        logger.info("accepted connection from: %s", client_address)
        return JsonConnection(connection, self.callback), client_address

    def reconnect(self):
        logger.debug("reconnecting...")
        try:
            self.close()
        except:
            pass
        try:
            self.bind()
            return True
        except:
            return False

    def close(self):
        try:
            os.unlink(self.server_address)
        except:
            pass
        super(self, JsonServerSocket).close()


class JsonConnection(JsonSocket):
    def __init__(self, connection, callback):
        self.callback = callback
        super(JsonConnection, self).__init__(connection)  # does not work

    def receive_call(self):
        request = self.receive()
        logger.debug("received request: %s", request)
        if 'quit' in request:
            self.close()
            return False
        function, args, kwargs = request['func'], request['args'], request['kwargs']
        try:
            ret = self.callback(function, args, kwargs)
            response = {'return': ret}
            logger.debug("sending response: %s", response)
            self.send(response)
            return True
        except Exception as e:
            response = {'exception': {"name": str(e.__class__), "msg": str(e)}}
            logger.debug("sending response: %s", response)
            self.send(response)
            return True


class JsonClientSocket(JsonSocket):
    def __init__(self, server_address):
        self.server_address = server_address
        super(JsonClientSocket, self).__init__(None)  # # does not work
        # super(JsonClientSocket, self).__init__(socket.socket(socket.AF_UNIX, socket.SOCK_STREAM))  # # does not work
        # self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

    def connect(self):
        # logger.debug("connecting to: %s", self.server_address)
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sock.connect(self.server_address)
        logger.debug("connected to: %s", self.server_address)

    def call_function(self, function, args=None, kwargs=None):
        request = {'func': function, 'args': args or list(), 'kwargs':kwargs or dict()}
        self.send(request)
        response = self.receive()
        logger.debug("received response: %s", response)
        # TODO: receive logs
        if 'exception' in response:
            _raise_exception(response['exception'])
        if 'return' in response:
            return response['return']

    def quit(self):
        request = {'quit': ''}
        self.send(request)
        self.close()

    def reconnect(self):
        logger.debug("reconnecting...")
        try:
            self.quit()
        except:
            pass
        try:
            self.connect()
            return True
        except:
            return False


class JsonServerMaster(threading.Thread):
    def __init__(self, server_address, threaded=False):
        self.server_address = server_address
        self.threaded = threaded
        self.server = JsonServerSocket(server_address, self.callback)
        self.up = True
        self.ipc_functions = dict()
        super().__init__()

    def ipc_function(self, f):
        self.ipc_functions[f.__name__] = f
        return f

    def callback(self, func, args, kwargs):
        try:
            func_call  = self.ipc_functions[func]
            return func_call(*args, **kwargs)
        except:
            logger.error("Error calling function", exc_info=True)
            raise

    def run(self):
        self.server.bind()
        try:
            while self.up:
                try:
                    connection, client_address = self.server.accept()
                    if self.up:
                        self._process_connection(connection, client_address)
                    else:
                        connection.close()
                except KeyboardInterrupt:
                        raise
                except Exception as e:
                    logger.error("Error in control server: %s", e, exc_info=True)
        finally:
            try:
                self.server.close()
            except:
                pass

    def _process_connection(self, connection, client_address):
        if self.threaded:
            t = threading.Thread(target=self._process_connection_run, args=(connection, client_address))
            t.setDaemon(True)
            t.start()
        else:
            self._process_connection_run(connection, client_address)

    def _process_connection_run(self, connection, client_address):
        try:
            while connection.receive_call():
                pass
        except KeyboardInterrupt:
            self.stop()
        except:
            logger.error("Error in connection", exc_info=True)
        finally:
            try:
                connection.close()
            except:
                pass

    def stop(self):
        if self.up:
            self.up =  False
            # os.kill(os.getpid(), signal.SIGINT)
            try:
                client = JsonClientSocket(self.server_address)
                client.connect()
                # client.quit()
            except:
                pass


class JsonServerProxy:
    def __init__(self, server_address):
        self.client = JsonClientSocket(server_address)
        self.client.connect()

    def __getattribute__(self, item):
        try:
            return super().__getattribute__(item)
        except:
            return self.get_func(item)

    def get_func(self, func):
        def _call_func(*args, **kwargs):
            return self.client.call_function(func, args, kwargs)
        return _call_func

    def quit(self):
        self.client.quit()


def _raise_exception(exception_object):
    exception_name, exception_msg = exception_object['name'], exception_object['msg']
    if exception_name == 'ValueError':
        raise ValueError(exception_msg)
    elif exception_name == 'KeyError':
        raise KeyError(exception_msg)
    else:
        raise Exception(exception_name + ": " + exception_msg)



import logging

logger = logging.getLogger(__name__)

ORGANIZATION_KEYS = ["org.name", "org-name", "orgname", "organization name", "organization name", "organisation",
                     "organization", "owner.name", "owner-name", "owner name", "owner",
                     "registrant organization",
                     "maintainer", "mnt-by", "name",
                     "registrant name", "registrar name", "registrar", "registrant"]

NETRANGE_KEYS = ["netrange", "inetnum", "inet6num"]
TLD_CHARSET = 'arpbthovieclgdumynsfkzxqwj-1432907865'


def whois(query):
    from easilyb.commands import run_command_ex1
    from easilyb.validators import validate_ip, validate_domain, validate_charset
    if not validate_ip(query) and not validate_domain(query) and not validate_charset(query, TLD_CHARSET):
        raise ValueError("Query is not an IP address nor a domain")
    ret_code, output = run_command_ex1(["whois", query])
    output = output.decode(errors="replace")
    output_parsed = _parse_whois_output(output)
    org = _find_organization(output_parsed)
    if org == '': org = None
    netrange = _find_netrange(output_parsed)
    if netrange == '': netrange = None
    emails = _find_emails(output)
    return org, netrange, emails, output


def cymru_whois(ip):
    from easilyb.commands import run_command_ex1
    from easilyb.validators import validate_ip
    if not validate_ip(ip):
        raise ValueError("Param is not an IP address")
    ret_code, output = run_command_ex1(["whois", "-h", "whois.cymru.com", "-v " + ip])
    keys = list()
    for l in output.decode(errors="replace").splitlines():
        if ip in l:
            values = [v.strip() for v in l.split('|')]
            ret = {keys[i]:values[i] for i in range(min(len(keys), len(values)))}
            for k in ['AS', 'BGP Prefix', 'AS Name']:
                if ret[k] == 'NA':
                    ret[k] = None
            if ret['CC'] == '':
                ret['CC'] = None
            return ret
        elif "BGP Prefix" in l:
            keys = [k.strip() for k in l.split('|')]
    return None


def _find_organization(output_parsed):
    return _find_keys_value(output_parsed, ORGANIZATION_KEYS)


def _find_netrange(output_parsed):
    return _find_keys_value(output_parsed, NETRANGE_KEYS)


def _find_emails(output):
    import re
    finds = []
    for line in output.splitlines():
        match = re.findall(r'[\w\.-]+@[\w\.-]+', line)
        if isinstance(match, list):
            finds.extend(match)
    return finds


def _parse_whois_output(output):
    ret = list()
    for l in output.splitlines():
        if ':' in l:
            splits = l.split(':')
            key, value = splits[0].strip(), (':'.join(splits[1:])).strip()
            ret.append((key, value))
    return ret


def _find_keys_value(output_parsed, keys):
    for key in keys:
        for k, v in output_parsed:
            if k.lower() == key and len(v) > 0:
                return v
    return None



#!/usr/bin/env python
"""
urlnorm.py - URL normalisation routines

urlnorm normalises a URL by;
  * lowercasing the scheme and hostname
  * taking out default port if present (e.g., http://www.foo.com:80/)
  * collapsing the path (./, ../, etc)
  * removing the last character in the hostname if it is '.'
  * unquoting any %-escaped characters

Available functions:
  norms - given a URL (string), returns a normalised URL
  norm - given a URL tuple, returns a normalised tuple
  test - test suite

CHANGES:
0.92 - unknown schemes now pass the port through silently
0.91 - general cleanup
     - changed dictionaries to lists where appropriate
     - more fine-grained authority parsing and normalisation
"""

__license__ = """
Copyright (c) 1999-2002 Mark Nottingham <mnot@pobox.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

__version__ = "0.93"


from urllib.parse import urlparse, urlunparse, unquote
try:
    lower = str.lower
except:
    from string import lower
import re

_collapse = re.compile('([^/]+/\.\./?|/\./|//|/\.$|/\.\.$)')
_server_authority = re.compile('^(?:([^\@]+)\@)?([^\:]+)(?:\:(.+))?$')
_default_port = {   'http': '80',
                    'https': '443',
                    'gopher': '70',
                    'news': '119',
                    'snews': '563',
                    'nntp': '119',
                    'snntp': '563',
                    'ftp': '21',
                    'telnet': '23',
                    'prospero': '191',
                }
_relative_schemes = [   'http',
                        'https',
                        'news',
                        'snews',
                        'nntp',
                        'snntp',
                        'ftp',
                        'file',
                        ''
                    ]
_server_authority_schemes = [   'http',
                                'https',
                                'news',
                                'snews',
                                'ftp',
                            ]


def norms(urlstring):
    """given a string URL, return its normalised form"""
    return urlunparse(norm(urlparse(urlstring)))


def norm(urltuple):
    """given a six-tuple URL, return its normalised form"""
    (scheme, authority, path, parameters, query, fragment) = urltuple
    scheme = lower(scheme)
    if authority:
        userinfo, host, port = _server_authority.match(authority).groups()
        if host[-1] == '.':
            host = host[:-1]
        authority = lower(host)
        if userinfo:
            authority = "%s@%s" % (userinfo, authority)
        if port and port != _default_port.get(scheme, None):
            authority = "%s:%s" % (authority, port)
    if scheme in _relative_schemes:
        last_path = path
        while 1:
            path = _collapse.sub('/', path, 1)
            if last_path == path:
                break
            last_path = path
    path = unquote(path)
    return (scheme, authority, path, parameters, query, fragment)

def norms_ex1(url):
    scheme, authority, path, parameters, query, fragment, userinfo, host, port = norm_ex1(urlparse(url))
    urltuple = scheme, authority, path, parameters, query, fragment
    url = urlunparse((scheme, authority, path, parameters, query, fragment))
    return urltuple, userinfo, host, port, url


def norm_ex1(urltuple):
    """given a six-tuple URL, return its normalised form"""
    (scheme, authority, path, parameters, query, fragment) = urltuple
    scheme = lower(scheme)
    if authority:
        userinfo, host, port = _server_authority.match(authority).groups()
        if host[-1] == '.':
            host = host[:-1]
        authority = lower(host)
        if userinfo:
            authority = "%s@%s" % (userinfo, authority)
        if port and port != _default_port.get(scheme, None):
            authority = "%s:%s" % (authority, port)
    if scheme in _relative_schemes:
        last_path = path
        while 1:
            path = _collapse.sub('/', path, 1)
            if last_path == path:
                break
            last_path = path
    path = unquote(path)
    return (scheme, authority, path, parameters, query, fragment, userinfo, host, port)



def test():
    """ test suite; some taken from RFC1808. """
    tests = {
        '/foo/bar/.':                    '/foo/bar/',
        '/foo/bar/./':                   '/foo/bar/',
        '/foo/bar/..':                   '/foo/',
        '/foo/bar/../':                  '/foo/',
        '/foo/bar/../baz':               '/foo/baz',
        '/foo/bar/../..':                '/',
        '/foo/bar/../../':               '/',
        '/foo/bar/../../baz':            '/baz',
        '/foo/bar/../../../baz':         '/../baz',
        '/foo/bar/../../../../baz':      '/baz',
        '/./foo':                        '/foo',
        '/../foo':                       '/../foo',
        '/foo.':                         '/foo.',
        '/.foo':                         '/.foo',
        '/foo..':                        '/foo..',
        '/..foo':                        '/..foo',
        '/./../foo':                     '/../foo',
        '/./foo/.':                      '/foo/',
        '/foo/./bar':                    '/foo/bar',
        '/foo/../bar':                   '/bar',
        '/foo//':                        '/foo/',
        '/foo///bar//':                  '/foo/bar/',
        'http://www.foo.com:80/foo':     'http://www.foo.com/foo',
        'http://www.foo.com:8000/foo':   'http://www.foo.com:8000/foo',
        'http://www.foo.com./foo/bar.html': 'http://www.foo.com/foo/bar.html',
        'http://www.foo.com.:81/foo':    'http://www.foo.com:81/foo',
        'http://www.foo.com/%7ebar':     'http://www.foo.com/~bar',
        'http://www.foo.com/%7Ebar':     'http://www.foo.com/~bar',
        'ftp://user:pass@ftp.foo.net/foo/bar': 'ftp://user:pass@ftp.foo.net/foo/bar',
        'http://USER:pass@www.Example.COM/foo/bar': 'http://USER:pass@www.example.com/foo/bar',
        'http://www.example.com./':      'http://www.example.com/',
        '-':                             '-',
    }

    n_correct, n_fail = 0, 0
    test_keys = list(tests.keys())
    test_keys.sort()
    for i in test_keys:
        print('ORIGINAL:', i)
        cleaned = norms(i)
        answer = tests[i]
        print('CLEANED: ', cleaned)
        print('CORRECT: ', answer)
        if cleaned != answer:
            print("*** TEST FAILED")
            n_fail = n_fail + 1
        else:
            n_correct = n_correct + 1
        print("")
    print("TOTAL CORRECT:", n_correct)
    print("TOTAL FAILURE:", n_fail)


if __name__ == '__main__':
    test()



import requests
import subprocess

from easilyb.queueworker import QueueWorker


import logging
logger = logging.getLogger(__name__)


USERAGENT = "Firefox"

def _request(callback, index, method, url, arg1, arg2, kwargs):
    logger.debug("requesting %s(%s) for %s"%(method, url, str(callback)))
    if method=='get':
        resp = requests.get(url,arg1,**kwargs)
    elif method=='post':
        resp = requests.post(url, arg1, arg2, **kwargs)
    callback(resp, index)

def _wget_wraper(url, fileoutput, referer=None, useragent=USERAGENT):
    cmd = ['wget', url, '-O', fileoutput, '-U', useragent]
    if referer: cmd+= ['--referer', referer]
    try: return subprocess.Popen(cmd).wait() #return subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE).wait()
    except:
        return 999

def _wget(url, fileoutput, callback=None, index=None, referer=None, useragent=USERAGENT):
    logger.debug("wgeting %s(%s) for %s"%(url, str(callback)))
    resp = _wget_wraper(url, fileoutput, referer, useragent)
    callback(resp, index)

class Requester:
    def __init__(self, threads=10, queue=None):
        self.queue = QueueWorker(_request, threads=threads, log_progress=False, queue=queue)

    def get(self, callback, url, index=None, params=None, **kwargs):
        self.queue.put((callback, index, 'get', url, params,None,kwargs))

    def post(self, callback, url, index=None, data=None, json=None, **kwargs):
        self.queue.put((callback, index, 'post', url, data, json, kwargs))
    def join(self):
        return self.queue.join()

class Wgeter:
    def __init__(self, threads=10, queue=None):
        self.queue = QueueWorker(_wget, threads=threads, log_progress=False, queue=queue)

    def wget(self, url, fileoutput, callback=None, index=None, referer=None, useragent=USERAGENT):
        self.queue.put((url, fileoutput, callback, index, referer, useragent))

    def join(self):
        return self.queue.join()


from threading import Lock
import os
import easilyb.json_serialize as jsons
import json
import yaml


def _load_json(path, driver=jsons):
    with open(path) as fin:
        return driver.load(fin)


def _save_json(data, path, indent=None, driver=jsons):
    with open(path, 'w') as fou:
        if driver == yaml:
            driver.dump(data, fou, indent=indent, default_flow_style=False)
        else:
            driver.dump(data, fou, indent=indent)


class JsonMinConnexion:
    """Minimalistic Json Database Connexion class."""

    def __init__(self, path, create=True, template=None, template_file=None, indent=3, readonly=False, driver="jsons"):
        """JsonMinDb constructor.

        :param path: json file path.
        :type path: str.
        :param create: Create if json file does not exist.
        :type create: bool.
        :param template: if create is True, create from template.
        :type template: dict.
        :param template_file: if create is True and template is None, create from template_file.
        :type template_file: str.
        :param indent: if create is True, create from template.
        :type indent: int.
        :raises: ValueError

        """
        self.path = path
        self.indent = indent
        self.readonly = readonly
        self.lock = Lock()
        if driver == "json":
            self.driver = json
        elif driver == "yaml":
            self.driver = yaml
        elif driver == "jsons" or driver == "json_serialize":
            self.driver = jsons
        else:
            raise ValueError("Unkown driver: %s!" % driver)
        if not os.path.isfile(path):
            if create:
                if template:
                    _template = template
                elif template_file:
                    _template = _load_json(template_file, driver=self.driver)
                else:
                    _template = {}

                _save_json(_template, path, indent=indent, driver=self.driver)
                self.db = _load_json(path, driver=self.driver)
            else:
                raise ValueError("Database file doesn't exist: " + path)
        self.db = _load_json(path, driver=self.driver)

        # db dict calls
        # self.__contains__ = self.db.__contains__
        # self.__delitem__ = self.db.__delitem__
        # self.__getitem__ = self.db.__getitem__
        # self.__iter__ = self.db.__iter__
        # self.__len__ = self.db.__len__
        # self.__setitem__ = self.db.__setitem__
        #
        # #self.clear = self.db.clear
        # #self.copy = self.db.copy
        # #self.fromkeys = self.db.fromkeys
        # try: # only in python2
        #     self.has_key = self.db.has_key
        #     self.iteritems = self.db.iteritems
        #     self.iterkeys = self.db.iterkeys
        #     self.itervalues = self.db.itervalues
        # except:
        #     pass
        # self.items = self.db.items
        # self.keys = self.db.keys
        # self.update = self.db.update
        # self.values = self.db.values

    def __contains__(self, k):
        return self.db.__contains__(k)

    def __delitem__(self, k):
        return self.db.__delitem__(k)

    def __getitem__(self, k):
        return self.db.__getitem__(k)

    def __iter__(self):
        return self.db.__iter__()

    def __len__(self):
        return self.db.__len__()

    def __setitem__(self, k, o):
        return self.db.__setitem__(k, o)

    def items(self):
        return self.db.items()

    def keys(self):
        return self.db.keys()

    def update(self):
        return self.db.update()

    def values(self):
        return self.db.values()

    def has_key(self):
        try: return self.db.has_key()
        except: pass

    def save(self):
        """updates database persistance file in the disk. """
        if self.readonly:
            raise Exception("Read Only Access!")
        _save_json(self.db, self.path, indent=self.indent, driver=self.driver)

    def reload(self):
        """reload database from disk."""
        self.db = _load_json(self.path, driver=self.driver)

    def __str__(self):
        return "<kutils.json_min_db.JsonMinConnexion instance %s>" % self.db.__str__()

    def __enter__(self):
        self.lock.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            if exc_type is None:
                self.save()
        finally:
            self.lock.release()



import re


def validate_charset(line, charset):
    for c in line:
        if not c in charset:
            return False
    return True


# -----------------------------------------------------------------------------------
# Credit: Konsta Vesterinen (github.com/kvesteri) LICENSE: MIT
DOMAIN_PATTERN = re.compile(
    r'^(([a-zA-Z]{1})|([a-zA-Z]{1}[a-zA-Z]{1})|'
    r'([a-zA-Z]{1}[0-9]{1})|([0-9]{1}[a-zA-Z]{1})|'
    r'([a-zA-Z0-9][-_.a-zA-Z0-9]{0,61}[a-zA-Z0-9]))\.'
    r'([a-zA-Z]{2,13}|[a-zA-Z0-9-]{2,30}.[a-zA-Z]{2,3})$'
)
# The MIT License (MIT)
#
# Copyright (c) 2013-2014 Konsta Vesterinen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# -----------------------------------------------------------------------------------


def validate_domain(domain):
    if DOMAIN_PATTERN.match(domain) is not None:
       return True
    return False

from easilyb.validators.ip import validate_ip, validate_ipv4, validate_ipv6




import json
import importlib
import codecs
import logging
from easilyb.cache import SimpleCache

_cache = SimpleCache(expires=False)
logger = logging.getLogger(__name__)
PRINTABLE = b'0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!"#$%&()*+,-./:;<=>?@[]^_`{|}~ '


class _BaseJsonSerializable(object):
    pass


@_cache.cached
def get_serializable_class(meta_prefix="_", ignore_local_vars=True):
    class _ParamJsonSerializable(_BaseJsonSerializable):
        def to_json_dict(self):
            return {k:self.__dict__[k] for k in self.__dict__ if not k.startswith(meta_prefix)
                    and not (ignore_local_vars and k.startswith('_')) and not k == 'to_json_dict'}

        def __getitem__(self, item):
            return self.__dict__.__getitem__(item)

        def __setitem__(self, key, value):
            return self.__dict__.__setitem__(key, value)

        def __contains__(self, item):
            return self.__dict__.__contains__(item)

        def __delitem__(self, key):
            return self.__dict__.__delitem__(key)

    return _ParamJsonSerializable


@_cache.cached
def get_encoder_class(meta_prefix="_", ignore_local_vars=True):
    _module = meta_prefix + 'module'
    _class = meta_prefix + 'class'
    _serializable = get_serializable_class(meta_prefix, ignore_local_vars)

    class _ParamJsonSerializeEncoder(json.JSONEncoder):
        def default(self, obj):
            try:
                if isinstance(obj, _BaseJsonSerializable):
                    # m = obj.__class__.__module__
                    # c = obj.__class__.__name__
                    # json_dict = obj.to_json_dict()
                    json_dict = _serializable.to_json_dict(obj)
                    json_dict[_module] = obj.__class__.__module__
                    json_dict[_class] = obj.__class__.__name__
                    # return {
                    #     "_module": m,
                    #     "_class": c,
                    #     "_value": obj.to_json_dict()
                    # }
                    return json_dict
                if isinstance(obj, bytes):
                    json_dict = {_class: 'bytes', _module:None}
                    try:
                        json_dict['utf8'] = obj.decode('utf8')
                    except:
                        json_dict['repr'] = repr_encode(obj)
                    return json_dict
                if isinstance(obj, set): # or isinstance(obj, tuple):
                    return {_class:obj.__class__.__name__, _module: None, 'lst': list(obj)}
                return json.JSONEncoder.default(self, obj)
            except:
                logger.error('Error serializing Json Object: %s; returning default serialization.', obj, exc_info=True)
                return json.JSONEncoder.default(self, obj)

    return _ParamJsonSerializeEncoder


@_cache.cached
def get_decoder_class(meta_prefix="_", ignore_local_vars=True):
    _module = meta_prefix + 'module'
    _class = meta_prefix + 'class'
    _serializable = get_serializable_class(meta_prefix, ignore_local_vars)

    class _ParamJsonSerializeDecoder(json.JSONDecoder):
        def __init__(self, *args, **kwargs):
            json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)
    
        def object_hook(self, obj):
            # if '_value' not in obj or _module not in obj or _class not in obj:
            try:
                if _module not in obj or _class not in obj:
                    return obj
                if obj[_module] is None:
                    if obj[_class] == 'bytes':
                        if 'utf8' in obj:
                            return obj['utf8'].encode()
                        else:
                            return repr_decode(obj['repr']) # security risk
                    elif obj[_class] == 'set':
                        return set(obj['lst'])
                    # elif obj[_class] == 'tuple':
                    #     return tuple(obj['lst'])
                    else:
                        return obj
                module_ = importlib.import_module(obj[_module])
                class_ = module_.__dict__[obj[_class]]
                obj_deser = _serializable()
                obj_deser.__module__ = obj[_module]
                obj_deser.__class__ = class_
                obj_deser.__dict__ = {k:obj[k] for k in obj if not k.startswith(meta_prefix) and not k == 'to_json_dict'}
                # return class_.from_json_dict(obj)
                return obj_deser
            except:
                logger.error('Error deserializing Json Object; returning dict: %s', obj, exc_info=True)
                return obj
    return _ParamJsonSerializeDecoder


JsonSerializable = get_serializable_class()
JsonSerializeEncoder = get_encoder_class()
JsonSerializeDecoder = get_decoder_class()


def load(*args, **kwargs):
    meta_prefix = kwargs.pop("meta_prefix", "_")
    ignore_local_vars = kwargs.pop("ignore_local_vars", True)
    if not 'cls' in kwargs or kwargs['cls'] is None:
        kwargs['cls'] = get_decoder_class(meta_prefix, ignore_local_vars)
    return json.load(*args, **kwargs)


def loads(*args, **kwargs):
    meta_prefix = kwargs.pop("meta_prefix", "_")
    ignore_local_vars = kwargs.pop("ignore_local_vars", True)
    if not 'cls' in kwargs or kwargs['cls'] is None:
        kwargs['cls'] = get_decoder_class(meta_prefix, ignore_local_vars)
    return json.loads(*args, **kwargs)


def dump(*args, **kwargs):
    meta_prefix = kwargs.pop("meta_prefix", "_")
    ignore_local_vars = kwargs.pop("ignore_local_vars", True)
    if not 'cls' in kwargs or kwargs['cls'] is None:
        kwargs['cls'] = get_encoder_class(meta_prefix, ignore_local_vars)
    return json.dump(*args, **kwargs)


def dumps(*args, **kwargs):
    meta_prefix = kwargs.pop("meta_prefix", "_")
    ignore_local_vars = kwargs.pop("ignore_local_vars", True)
    if not 'cls' in kwargs or kwargs['cls'] is None:
        kwargs['cls'] = get_encoder_class(meta_prefix, ignore_local_vars)
    return json.dumps(*args, **kwargs)


def repr_encode(bin):
    res = u''
    for ci in bin:
        if isinstance(ci, int):
            c = bytes([ci])
        else:
            c = ci
        if c in PRINTABLE:
            res += c.decode('utf8')
        else:
            # hex_ = hex(c)[2:]
            # if len(c) == 1:
            #     res += u'\\0'+ hex_
            # else:
            #     res += u'\\'+ hex_
            res += u'\\' + codecs.encode(c, "hex_codec").decode('utf8')
    return res


def repr_decode(repr_):
    res = b''
    i = 0
    while i < len(repr_):
        if repr_[i] != u'\\':
            res += repr_[i].encode('utf8')
            i += 1
        else:
            res += codecs.decode(repr_[i+1:i+3], "hex_codec")
            i += 3
    return res




import os


def pidlock(lock_path):
    lock_held = False
    if os.path.exists(lock_path):
        try:
            with open(lock_path) as fin:
                pid = fin.readline().strip()
                if os.path.exists("/proc/%s" % pid):
                    lock_held = True
                else:
                    os.unlink(lock_path)
        except:
            os.unlink(lock_path)
    if lock_held:
        raise LockHeld()

    with open(lock_path, 'w') as fou:
        fou.write("%d" % os.getpid())


def release_lock(lock_path):
    os.unlink(lock_path)


class LockHeld(Exception):
    pass


from time import time as _time
from threading import Lock as _Lock
from easilyb.string_ops import random_secure_string

import logging
logger = logging.getLogger(__name__)

MAX_CSRF_TOKEN_AGE = 600 # 10min
PURGE_RATE = 1200 # 20min
_mapping = None


class _CSRFMapping:
    def __init__(self):
        self._mapping = dict()
        self._purge_limit = _time() + MAX_CSRF_TOKEN_AGE
        self._lock = _Lock()

    def insert(self, token, form_id, referer, session_id, tmax):
        try:
            self._lock.acquire()
            self._insert(token, form_id, referer, session_id, tmax)
        finally:
            self._lock.release()

    def _insert(self, token, form_id, referer, session_id, tmax):
        t = _time()
        if t > self._purge_limit:
            self._purge()
        self._mapping[token] = (form_id, referer, session_id, t, tmax)


    def check(self, token, form_id, referer, session_id, unvalidate):
        try:
            self._lock.acquire()
            return self._check(token, form_id, referer, session_id,  unvalidate)
        finally:
            self._lock.release()

    def _check(self, token, form_id, referer, session_id, unvalidate):
        t = _time()
        if t > self._purge_limit:
            self._purge()
        if token in self._mapping:
            form_id0, referer0, session_id0, t0, tmax0 = self._mapping[token]
            if (form_id0 is None or form_id0 == form_id) \
                    and (referer0 is None or referer0 == referer) \
                    and (session_id0 is None or session_id0 == session_id) \
                    and t < t0 + (MAX_CSRF_TOKEN_AGE if tmax0 is None else tmax0):
                if unvalidate:
                    del self._mapping[token]
                return True
        return False


    def purge(self):
        try:
            self._lock.acquire()
            self._purge()
        finally:
            self._lock.release()

    def _purge(self):
        _to_remove = list()
        t = _time()
        for token in self._mapping:
            form_id0, referer0, session_id0, t0, tmax = self._mapping[token]
            if t > t0 + (MAX_CSRF_TOKEN_AGE if tmax is None else tmax):
                _to_remove.append(token)
        for token in _to_remove:
            del self._mapping[token]

def _get_csrf_mapping():
    global _mapping
    if _mapping is None:
        _mapping = _CSRFMapping()
    return _mapping

def get_csrf_token(form_id=None, referer=None, session_id=None, tmax=None):
    mapping = _get_csrf_mapping()
    token = random_secure_string()
    mapping.insert(token, form_id, referer, session_id, tmax)
    return token

def check_csrf_token(token, form_id=None , referer=None, session_id=None, unvalidate=True):
    mapping = _get_csrf_mapping()
    return mapping.check(token, form_id, referer, session_id, unvalidate)






from urllib.parse import urlparse, urlunparse, parse_qs
from easilyb import urlnorm as un
import logging
logger = logging.getLogger(__name__)


class UrlSelector:
    def __init__(self):
        self.seen = set()
        self.notseen = set()

    def put(self, url):
        item = UrlItem(url)
        # check duplicates
        if item in self.seen or item in self.notseen:
            return
        self.notseen.add(item)

    def pop(self):
        if len(self.notseen) == 0:  # if not seen is empty
            return None
        if len(self.seen) == 0:  # if seen is empty (pick arbitrary element)
            for curitem in self.notseen:
                break
            self.notseen.remove(curitem)
            self.seen.add(curitem)
            return curitem.url_original

        curitem = None
        curdist = -1
        for n in self.notseen:
            sdist = -1
            for s in self.seen:
                dist = s.dist(n)
                if sdist == -1 or dist < sdist:
                    sdist = dist
            if sdist > curdist:
                curdist = sdist
                curitem = n

        self.notseen.remove(curitem)
        self.seen.add(curitem)
        return curitem.url_original

    def __len__(self):
        return len(self.notseen)


class UrlItem:
    def __init__(self, url):
        self.url_norm = un.norm_ex1(urlparse(url))
        self.url = urlunparse(self.url_norm[:-4] + ('',))
        try:
            self.url_unicode = self.url.decode('utf8')
        except AttributeError:
            self.url_unicode = self.url
        self.url_original = url
        self.host = self.url_norm[-2]
        self.distances = dict()

    def set_distance(self, item, dist):
        self.distances[item.url] = dist

    def get_distance(self, item):
        if item.url in self.distances:
            return self.distances[item.url]
        else:
            raise KeyError('UrlItem key does not exist: ' + str(item))

    def dist(self, item):
        if item.url in self.distances:
            return self.distances[item.url]
        dist = url_distance(self.url_norm, item.url_norm)
        self.set_distance(item, dist)
        item.set_distance(self, dist)
        return dist

    def __str__(self):
        return "UrlItem(%s)" % (self.url)

    def __eq__(self, other):
        return self.url == other.url

    def __hash__(self):
        return hash(self.url)


def url_distance(url1, url2):
    scheme1, authority1, path1, parameters1, query1, fragment1, userinfo1, host1, port1 = url1
    scheme2, authority2, path2, parameters2, query2, fragment2, userinfo2, host2, port2 = url2
    sdiff = servicediff(scheme1, authority1, host1, scheme2, authority2, host2)
    if sdiff > 0:
        return sdiff * 50
    pdiff = pathdiff(path1, path2)
    if pdiff > 0:
        return pdiff * 2
    return querrydiff(query1, query2)


def servicediff(scheme1, authority1, host1, scheme2, authority2, host2):
    if scheme1 + authority1 == scheme2 + authority2:
        return 0
    if host1 == host2:
        return 10
    return min(20, edit_dist(host1, host2)) + 10


def pathdiff(path1, path2):
    if path1 == path2: return 0
    return edit_dist(path1, path2)


def querrydiff(query1, query2):
    dict1 = parse_qs(query1)
    dict2 = parse_qs(query2)
    diff = 0
    for k in dict1:
        if k not in dict2:
            diff += 10
        else:
            diff += edit_dist(','.join(dict1[k]), ','.join(dict2[k]))
            del dict2[k]
    diff += 10 * len(dict2)
    return diff


def edit_dist(str1, str2):
    l1 = len(str1) + 1
    l2 = len(str2) + 1
    arr = {}

    for i in range(l1): arr[i, 0] = i
    for j in range(l2): arr[0, j] = j
    for i in range(1, l1):
        for j in range(1, l2):
            c = 0 if str1[i - 1] == str2[j - 1] else 1
            arr[i, j] = min(arr[i, j - 1] + 1, arr[i - 1, j] + 1, arr[i - 1, j - 1] + c)

    return arr[i, j]



def _get_ip_object(ip):
    from ipaddress import ip_address
    try:
        ip_object = ip_address(ip)
        return ip_object
    except:
        return None

def validate_ip(ip):
    if _get_ip_object(ip) is None:
        return False
    return True


def validate_ipv4(ip):
    obj = _get_ip_object(ip)
    if obj is None or obj.version != 4:
        return False
    return True


def validate_ipv6(ip):
    obj = _get_ip_object(ip)
    if obj is None or obj.version != 6:
        return False
    return True


import random
import string
from Crypto.Random import random as cryptorandom

import logging
logger = logging.getLogger(__name__)


def random_string(size=8, upper=True, lower=False, digits=True , secure=False):
    space = []
    if upper:
        space += string.ascii_uppercase
    if lower:
        space += string.ascii_lowercase
    if digits:
        space += string.digits
    if  len(space) == 0:
        raise ValueError("At least one of (upper, lower, digits) should be True!")
    if secure:
        _random = cryptorandom
    else:
        _random = random
    return ''.join(_random.choice(space) for i in range(size))


def random_secure_string(size=64):
    return random_string(size, upper=True, lower=True, digits=True , secure=True)



import logging
from easilyb.json_min_db import JsonMinConnexion
from threading import Lock

logger = logging.getLogger(__name__)
NOTHING = object()


class EasyConfig:
    def __init__(self, filepath, create=True, template=None, template_file=None, indent=3, driver="yaml"):
        self.db = JsonMinConnexion(filepath, create=create, template=template, template_file=template_file,
                                   indent=indent, driver=driver)
        self._lock = Lock()
        self._save = True

    def get(self, key, default=NOTHING, set_default=True, save=True):
        try:
            return self.db[key]
        except:
            if default == NOTHING:
                raise KeyError(key)
            else:
                if set_default:
                    self.set(key, default, save=save)
                return default

    def set(self, key, value, save=True):
        if self._save:
            with self._lock:
                self.db[key] = value
                if save:
                    self.db.save()
        else:
            self.db[key] = value

    def save(self):
        self.db.save()

    def __enter__(self):
        self._lock.acquire()
        self._save = False
        self.db.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.db.__exit__(exc_type, exc_val, exc_tb)
        finally:
            self._save = True
            self._lock.release()



from time import time as _time
try:
    import threading as _threading
except ImportError:
    import dummy_threading as _threading
from . import UrlItem
import logging
logger = logging.getLogger(__name__)

class RateQueue:
    """Create a queue object with a given maximum size.

    If maxsize is <= 0, the queue size is infinite.
    """
    def __init__(self, maxsize=0, maxrate=0.01, domainrate=None, limit=None):
        self.maxsize = maxsize
        self.mutex = _threading.Lock()
        self.not_empty = _threading.Condition(self.mutex)
        self.not_full = _threading.Condition(self.mutex)
        self.all_tasks_done = _threading.Condition(self.mutex)
        self.unfinished_tasks = 0
        self.seen = set()
        self.notseen = set()
        self.limit = limit
        self._count = 0
        self.maxrate = maxrate
        self.domainrate = domainrate
        self._lasttimes = dict()
        self._lasttime = -1


    def task_done(self):
        """Indicate that a formerly enqueued task is complete.

        Used by Queue consumer threads.  For each get() used to fetch a task,
        a subsequent call to task_done() tells the queue that the processing
        on the task is complete.

        If a join() is currently blocking, it will resume when all items
        have been processed (meaning that a task_done() call was received
        for every item that had been put() into the queue).

        Raises a ValueError if called more times than there were items
        placed in the queue.
        """
        self.all_tasks_done.acquire()
        try:
            unfinished = self.unfinished_tasks - 1
            if unfinished <= 0:
                if unfinished < 0:
                    raise ValueError('task_done() called too many times')
                self.all_tasks_done.notify_all()
            self.unfinished_tasks = unfinished
        finally:
            self.all_tasks_done.release()

    def join(self):
        """Blocks until all items in the Queue have been gotten and processed.

        The count of unfinished tasks goes up whenever an item is added to the
        queue. The count goes down whenever a consumer thread calls task_done()
        to indicate the item was retrieved and all work on it is complete.

        When the count of unfinished tasks drops to zero, join() unblocks.
        """
        self.all_tasks_done.acquire()
        try:
            while self.unfinished_tasks:
                self.all_tasks_done.wait()
        finally:
            self.all_tasks_done.release()

    def qsize(self):
        """Return the approximate size of the queue (not reliable!)."""
        self.mutex.acquire()
        n = self._qsize()
        self.mutex.release()
        return n

    def empty(self):
        """Return True if the queue is empty, False otherwise (not reliable!)."""
        self.mutex.acquire()
        n = not self._qsize()
        self.mutex.release()
        return n

    def full(self):
        """Return True if the queue is full, False otherwise (not reliable!)."""
        self.mutex.acquire()
        n = 0 < self.maxsize == self._qsize()
        self.mutex.release()
        return n

    def put(self, item, block=True, timeout=None):
        """Put an item into the queue.

        If optional args 'block' is true and 'timeout' is None (the default),
        block if necessary until a free slot is available. If 'timeout' is
        a positive number, it blocks at most 'timeout' seconds and raises
        the Full exception if no free slot was available within that time.
        Otherwise ('block' is false), put an item on the queue if a free slot
        is immediately available, else raise the Full exception ('timeout'
        is ignored in that case).
        """
        self.not_full.acquire()
        try:
            if self.maxsize > 0:
                if not block:
                    if self._qsize() == self.maxsize:
                        raise Full
                elif timeout is None:
                    while self._qsize() == self.maxsize:
                        self.not_full.wait()
                elif timeout < 0:
                    raise ValueError("'timeout' must be a positive number")
                else:
                    endtime = _time() + timeout
                    while self._qsize() == self.maxsize:
                        remaining = endtime - _time()
                        if remaining <= 0.0:
                            raise Full
                        self.not_full.wait(remaining)
            self._put(item)
            self.unfinished_tasks += 1
            self.not_empty.notify()
        finally:
            self.not_full.release()

    def put_nowait(self, item):
        """Put an item into the queue without blocking.

        Only enqueue the item if a free slot is immediately available.
        Otherwise raise the Full exception.
        """
        return self.put(item, False)

    def get(self, block=True, timeout=None):
        return self.pop(block, timeout)

    def pop(self, block=True, timeout=None):
        """Remove and return an item from the queue.

        If optional args 'block' is true and 'timeout' is None (the default),
        block if necessary until an item is available. If 'timeout' is
        a positive number, it blocks at most 'timeout' seconds and raises
        the Empty exception if no item was available within that time.
        Otherwise ('block' is false), return an item if one is immediately
        available, else raise the Empty exception ('timeout' is ignored
        in that case).
        """
        self.not_empty.acquire()
        try:
            if not block:
                if not self._qsize():
                    raise Empty
                mindists = self._calc_mindists()
                t1 = _time()
                if self._check_mindists(mindists, t1) is not None:
                    raise NoCandidate

            elif timeout is None:
                while not self._qsize():
                    self.not_empty.wait()
                while True:
                    mindists = self._calc_mindists()
                    t1 = _time()
                    wait = self._check_mindists(mindists, t1)
                    if wait  is None:
                        break
                    self.not_empty.wait(wait)

            elif timeout < 0:
                raise ValueError("'timeout' must be a positive number")
            else:
                endtime = _time() + timeout
                while not self._qsize():
                    remaining = endtime - _time()
                    if remaining <= 0.0:
                        raise Empty
                    self.not_empty.wait(remaining)
                while True:
                    mindists = self._calc_mindists()
                    t1 = _time()
                    wait = self._check_mindists(mindists, t1)
                    if wait  is None:
                        break
                    remaining = endtime - _time()
                    if remaining <= 0.0:
                        raise NoCandidate
                    self.not_empty.wait(min(wait,remaining))
            item = self._get(t1, mindists)
            self.not_full.notify()
            return item
        finally:
            self.not_empty.release()

    def get_nowait(self):
        """Remove and return an item from the queue without blocking.

        Only get an item if one is immediately available. Otherwise
        raise the Empty exception.
        """
        return self.get(False)


    def _qsize(self, len=len):
        return len(self.notseen)

    # Put a new item in the queue
    def _put(self, item):
        item = UrlItem(item)
        # check duplicates
        if item in self.seen or item in self.notseen:
            return
        self.notseen.add(item)

    # Get an item from the queue
    def _get(self, t1, mindists):

        curitem = None
        for url, dist in mindists:
            if self._checkrate(url, t1) is None:
                curitem = url
                self._updaterate(url, t1)
                self.notseen.remove(curitem)
                self.seen.add(curitem)
                break
        assert curitem is not None
        print(( self._count))
        if self.limit: #TODO: this code is not tested and have problems while threading
            self._count +=1
            if self._count >= self.limit:
                logger.info("Limit reached: %d", self._count)
                self.unfinished_tasks = 0
                self.seen = set()
                self.notseen = set()
                self.maxsize = 0

        return curitem.url_original

    def _calc_mindists(self):
        mindists = list()
        for n in self.notseen:
            sdist = -1
            for s in self.seen:
                dist = s.dist(n)
                if sdist == -1 or dist < sdist:
                    sdist = dist
            mindists.append((n, sdist))
        mindists.sort(key=lambda x:x[1], reverse=True)
        return mindists

    def _check_mindists(self, mindists, t1):
        minwait = -1
        for url, dist in mindists:
            wait = self._checkrate(url, t1)
            if wait is None:
                return None
            if minwait == -1 or wait < minwait:
                minwait = wait
        assert  minwait > -1
        return minwait

    def _checkrate(self, url, t1):
        if self.domainrate is None or url.host not in self._lasttimes or t1 - self._lasttimes[url.host] > self.domainrate:
            domainwait = -1
        else:
            domainwait = self.domainrate - (t1 - self._lasttimes[url.host])
        interval = t1 - self._lasttime
        if self.maxrate is None or interval > self.maxrate:
            maxwait = -1
        else:
            maxwait = self.maxrate - interval
        wait = max(domainwait, maxwait)
        if wait <=0:
            return None
        return wait


    def _updaterate(self, url, t1):
        self._lasttime = t1
        self._lasttimes[url.host] = t1

    def __len__(self):
        return self.qsize()

class Empty(Exception):
    "Exception raised by Queue.get(block=0)/get_nowait()."
    pass

class NoCandidate(Exception):
    "Exception raised by Queue.get(block=0)/get_nowait()."
    pass

class Full(Exception):
    "Exception raised by Queue.put(block=0)/put_nowait()."
    pass



import logging
logger = logging.getLogger(__name__)

def read_all(sock):
    sock.settimeout(5.0)
    data = ""
    while True:
        part = sock.recv(4096)
        data += part
        if len(part) < 4096:
            break
    return data


import os
from easilyb.commands import get_command_output, run_command

import logging
logger = logging.getLogger(__name__)

_PS_TO_INT = ['PID', 'PPID', 'RSS']



def meminfo():
    total, free= 0,0
    with open('/proc/meminfo') as fin:
        for line in fin.readlines():
            sline = line.split()
            if str(sline[0]) == 'MemTotal:':
                total = int(sline[1])
            elif str(sline[0]) == 'MemFree:':
                free = int(sline[1])
    return total, free




def get_processes():
    output, err = get_command_output(['ps', '-Ao', 'pid,ppid,uname,rss,comm=PROG,command=CMDLINE'])
    lines = output.split('\n')
    header = lines[0].split()
    processes = [line.split(None, len(header) - 1) for line in lines[1:] if len(line)>0]
    processes2 = list()
    for processe in processes:
        processes2.append({header[j]: (int(processe[j]) if header[j] in _PS_TO_INT else processe[j])for j in range(len(header))})
    return processes2

def get_process(pid, processes=None):
    if processes is None: processes = get_processes()
    for process in processes:
        if process['PID']==pid:
            return process
    return None

def get_process_children(pid=None, recursive=False, processes=None):
    childs = list()
    if pid is None:
        pid = os.getpid()
    if processes is None: processes = get_processes()
    for p in processes:
        if int(p['PPID']) == pid:
            childs.append(p)
            if recursive:
                childs.extend(get_process_children(pid=p['PPID'], recursive=False, processes=processes))
    return childs


def get_open_ports():
    return_code, output = run_command(["netstat", "-lnt"])
    ret = list()
    if return_code != 0:
        logger.error("netstat returned a non zero code: %d", return_code)
    else:
        lines = output.split('\n')
        for line in lines:
            if 'LISTEN' in line and not 'tcp6' in line:
                try:
                    parts = line.split()
                    host, port = parts[3].split(':')
                    port = int(port)
                    if not port in ret:
                        ret.append(port)
                except Exception as e:
                    logger.warning("exception parsing a line in netstat output: %s (line=%s)" % (str(e), line), exc_info=True)
    return ret


from easilyb.urlselector import  UrlSelector
from easilyb.net.requestqueue import Requester
from lxml.html import fromstring
import logging
logger = logging.getLogger(__name__)



def _crawler_callback(resp, index=None):
    url,counter, depth, crawler = index
    crawler._parse_response(url, depth, resp)


class LxmlXpathLinkExtractor:
    def __init__(self):
        pass
    def extract_links(self, baseurl, response):
        html = fromstring(response, base_url=baseurl)
        html.make_links_absolute()
        return html.xpath('//a/@href')

class Crawler:
    def __init__(self, baseurls, callback, urlselector=None, urlextractor=None, maxdepth=3, maxurls=100, threads=4):
        self.baseurls = baseurls
        self.callback = callback
        self._urlselector = urlselector or UrlSelector()
        self._urlextractor = urlextractor or LxmlXpathLinkExtractor()
        self.maxdepth = maxdepth
        self.maxurls = maxurls #TODO: can't with the current workerqueue or urlselector implementation
        self._urlselector.limit = maxurls #TODO: may be dangerous
        self._threads = threads
        self._requester = Requester(threads=self._threads, queue=self._urlselector)
        self._counter = 0
        for u in self.baseurls:
            index = (u,self._counter,  0, self,)
            self._requester.get(_crawler_callback, u, index)
            self._counter+=1

    def _parse_response(self, url, depth, response):
        self.callback(url, response)
        if depth + 1 < self.maxdepth:
            logger.debug('Parsing response for url=%s', url)
            links = self._urlextractor.extract_links(url, response.text)
            for u in links:
                logger.debug("Found link: %s", u)
                index = (u,self._counter,  depth + 1, self,)
                self._requester.get( _crawler_callback, u, index)
                self._counter+=1

    def join(self):
        self._requester.join()










import socket
import logging
logger = logging.getLogger(__name__)


def get_tcp_banner(ip, port):
    try:
        s = socket.socket()
        s.settimeout(5.0)
        s.connect((ip, int(port)))
        s.send(b'GET /\n\n')
        banner = s.recv(100)
        try:
            banner_dec = banner.decode()
        except:
            banner_dec = repr(banner)[2:-1]
        return banner_dec
    except:
        logger.error("Error while getting TCP banner for %s.", (ip, port,), exc_info=True)
        return ''
    finally:
        try: s.close()
        except: pass


def get_udp_banner(ip, port):
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.settimeout(5.0)
        s.connect((ip, int(port)))
        s.send(b'GET /\n\n')
        banner = s.recv(100)
        try:
            banner_dec = banner.decode()
        except:
            banner_dec = repr(banner)[2:-1]
        return banner_dec
    except:
        logger.error("Error while getting UDP banner for %s.", (ip, port,), exc_info=True)
        return ''
    finally:
        try: s.close()
        except: pass


def getdomainsbyip(ip):
    try:
        ret = socket.gethostbyaddr(ip)
        return [ret[0]] + ret[1]
    except:
        return []


def getipsbydomain(domain):
    try:
        return socket.gethostbyname_ex(domain)[2]
    except:
        return []


import subprocess
import os
import time
import shlex
import json
from threading import Lock

import logging
logger = logging.getLogger(__name__)


DEFAULT_CMD_LOGFILE = "external_commands.log"
CMD_LOG_LOCK = Lock()


def run_command(command_line, log=False, log_file=None):
    if not isinstance(command_line, list):
        command_line = shlex.split(str(command_line))


    try:
        process = subprocess.Popen(command_line, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, )
        t0 = time.time()
        output, stderrdata =  process.communicate()
        return_code = process.poll()
        t1 = time.time()

        if log:
            if log_file is None:
                log_file = DEFAULT_CMD_LOGFILE
            mode = 'w'
            if os.path.isfile(log_file):
                mode = 'a'
            with CMD_LOG_LOCK:
                with open(log_file, mode) as fou:
                    #cmd_line = ' '.join(shlex.quote(x) for x in split_command)
                    fou.write(json.dumps({'cmd_line': ' '.join(command_line), 'return_code': return_code, 'output':output, 'start_time':t0, 'finish_time':t1})+'\n')
        return return_code, output
    except Exception as e:
        raise e


def run_command_ex1(command_line, log=False, log_file=None, yield_=False, timeout=None):

    if not isinstance(command_line, list):
        command_line = shlex.split(str(command_line))
    if isinstance(timeout, int) and timeout > 0:
        command_line = ['timeout', str(timeout)] + command_line
    logger.debug("Runinnng command: %s", command_line)

    process = subprocess.Popen(command_line, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, )
    t0 = time.time()
    output, stderrdata =  process.communicate()
    return_code = process.poll()
    t1 = time.time()

    if log:
        if log_file is None:
            log_file = DEFAULT_CMD_LOGFILE
        mode = 'w'
        if os.path.isfile(log_file):
            mode = 'a'
        with CMD_LOG_LOCK:
            with open(log_file, mode) as fou:
                #cmd_line = ' '.join(shlex.quote(x) for x in split_command)
                fou.write(json.dumps({'cmd_line': ' '.join(command_line), 'return_code': return_code, 'output':output, 'start_time':t0, 'finish_time':t1, 'timeout':timeout})+'\n')
    return return_code, output


def get_command_output(cmd_vector):
    p = subprocess.Popen(cmd_vector, stdout=subprocess.PIPE)
    return p.communicate()











import queue as Queue
import threading
import logging
import time

logger = logging.getLogger(__name__)


def _worker_wraper(i, queue_worker):
    while True:
        try:
            arg = queue_worker.get()
            logger.debug('_worker wraper i=%s arg=%s', i, arg)
            queue_worker.workerfunc(*arg)
        except KeyboardInterrupt:
            logger.info('Keyboard Interrupt', exc_info=True)
        except Exception:
            logger.exception('Exception while executing %s(%s)', str(queue_worker.workerfunc).split()[1], str(arg))
        finally:
            queue_worker.task_done()
            queue_worker.log_progress()


class QueueWorker:
    def __init__(self, workerfunc, queue=None, threads=2, log_progress=True, progress_interval=60):
        self.workerfunc = workerfunc
        self.threads = threads
        if queue is not None:
            self.queue = queue
        else:
            self.queue = Queue.Queue()
        self._init_threads()
        self.log_progress_ = log_progress
        self.progress_interval = progress_interval
        self.progress_lock = threading.Lock()
        self.lastprogress = 0
        self.total_tasks = 0

    def _init_threads(self):
        for i in range(self.threads):
            worker = threading.Thread(target=_worker_wraper, args=(i, self))
            worker.setDaemon(True)
            worker.start()

    def log_progress(self):
        if self.log_progress_:
            self.progress_lock.acquire()
            t1 = int(time.time())
            if t1 - self.lastprogress > self.progress_interval:
                logger.info('Progress: %d / %d done (%d%%)', self.queue.qsize(), self.total_tasks,
                            (100 * self.queue.qsize() / self.total_tasks))  # TODO: Log or print?
                self.lastprogress = t1
            self.progress_lock.release()

    def empty(self):
        return self.queue.empty()

    def full(self):
        return self.queue.full()

    def get(self, block=True, timeout=None):
        return self.queue.get(block=block, timeout=timeout)

    def get_nowait(self):
        return self.queue.get_nowait()

    def join(self):
        return self.queue.join()

    def put(self, item, block=True, timeout=None):
        self.total_tasks += 1
        return self.queue.put(item, block=block, timeout=timeout)

    def put_nowait(self, item):
        self.total_tasks += 1
        return self.queue.put_nowait(item)

    def put_bulk(self, lst):
        for item in lst:
            self.queue.put(item)
        self.total_tasks += len(lst)

    def qsize(self):
        return self.queue.qsize()

    def task_done(self):
        return self.queue.task_done()



#!/usr/bin/env python3

from distutils.core import setup

import os


def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            if not path.endswith('__pycache__') and not filename.endswith(".pyc"):
                paths.append(os.path.relpath(os.path.join(path, filename), directory))
    return paths

extra_files = package_files('easilyb/')

#print extra_files

setup(
      name='easilyb',
      version='0.12.0', # major.minor.fix: MAJOR incompatible API changes, MINOR add backwards-compatible functionality, FIX bug fixes
      description='Frequently used functions library for Python3 By Khalid Grandi (github.com/xaled).',
      long_description='Frequently used functions library for Python3 By Khalid Grandi (github.com/xaled).',
      long_description_content_type='text/x-rst',
      keywords='library utils common',
      author='Khalid Grandi',
      author_email='kh.grandi@gmail.com',
      url='https://github.com/xaled/easilyb/',
      install_requires=['requests', 'pycrypto', 'pyaml', 'lxml'],
      python_requires='>=3',
      packages=['easilyb'],
      package_data={'': extra_files},
     )



import time
import logging
import json
from threading import Lock
logger = logging.getLogger(__name__)
DEFAULT_TTL = 3600
DEFAULT_NAMESPACE = "DEFAULT_NAMESPACE"
DEFAULT_CLEAN_INTERVAL = 60


class SimpleCache:
    def __init__(self, clean_interval=DEFAULT_CLEAN_INTERVAL, expires=True, ttl=DEFAULT_TTL):
        self._db = dict()
        self.clean_interval = clean_interval
        self._last_clean = time.time()
        self._lock = Lock()
        self.expires = expires
        self.ttl = ttl

    def get(self, key, action=None, action_kwargs=None, namespace=DEFAULT_NAMESPACE,
            expires=None, ttl=None):
        if expires is None: expires = self.expires
        if ttl is None: ttl = self.ttl
        with self._lock:
            self._auto_clean()
            if namespace not in self._db:
                self._db[namespace] = dict()
            if key in self._db[namespace] and not self._db[namespace][key].expired():
                return self._db[namespace][key].value
            elif action is not None:
                action_kwargs = action_kwargs or dict()
                ret = action(**action_kwargs)
                entry = CacheEntry(key, ret, expires=expires, ttl=ttl, action=action, action_kwargs=action_kwargs)
                self._db[namespace][key] = entry
                return self._db[namespace][key].value
            else:
                raise NotFoundInCache()

    def put(self, key, value, action=None, action_kwargs=None, namespace=DEFAULT_NAMESPACE,
            expires=None, ttl=None):
        if expires is None: expires = self.expires
        if ttl is None: ttl = self.ttl
        with self._lock:
            self._auto_clean()
            if namespace not in self._db:
                self._db[namespace] = dict()
            entry = CacheEntry(key, value, expires=expires, ttl=ttl, action=action, action_kwargs=action_kwargs)
            self._db[namespace][key] = entry

    def cached(self, func=None, expires=None, ttl=None, namespace=DEFAULT_NAMESPACE):
        def decorator(func):
            def wrapper(*a, **ka):
                expires_ = self.expires if expires is None else expires
                ttl_ = self.ttl if ttl is None else ttl
                namespace_ = namespace + "." + func.__name__
                key = _get_key(a, ka)
                try:
                    return self.get(key, namespace=namespace_)
                except NotFoundInCache:
                    value = func(*a, **ka)
                    self.put(key, value, expires=expires_, ttl=ttl_, namespace=namespace_)
                    return value
            return wrapper

        if callable(func):
            return decorator(func)
        else:
            return decorator

    def get_namespace_cache(self, namespace, expires=None, ttl=None):
        if expires is None: expires = self.expires
        if ttl is None: ttl = self.ttl
        return NamespaceCache(self, namespace, expires=expires, ttl=ttl)

    def _auto_clean(self):
        if time.time() - self._last_clean > self.clean_interval:
            self.clean()

    def clean(self):
        for sp in self._db:
            to_remove = list()
            for k,v in self._db[sp].items():
                if v.expired():
                    to_remove.append(k)
            for k in to_remove:
                del self._db[sp][k]
        self._last_clean = time.time()


class NamespaceCache:
    def __init__(self, maincache, namespace, expires=True, ttl=DEFAULT_TTL):
        self.namespace = namespace
        self.cache = maincache
        self.expires = expires
        self.ttl = ttl

    def get(self, key, action=None, action_kwargs=None, expires=None, ttl=None):
        if expires is None: expires = self.expires
        if ttl is None: ttl = self.ttl
        return self.cache.get(key, action=action, action_kwargs=action_kwargs, expires=expires, ttl=ttl,
                       namespace=self.namespace)

    def put(self, key, value, action=None, action_kwargs=None, expires=None, ttl=None):
        if expires is None: expires = self.expires
        if ttl is None: ttl = self.ttl
        self.cache.put(key, value, action=action, action_kwargs=action_kwargs, expires=expires, ttl=ttl,
                       namespace=self.namespace)

    def cached(self, func=None, expires=None, ttl=None):
        if expires is None: expires = self.expires
        if ttl is None: ttl = self.ttl
        return self.cache.cached(func, expires, ttl, namespace=self.namespace)


class NotFoundInCache(Exception):
    pass


class CacheEntry:
    def __init__(self, key, value, expires=True, ttl=DEFAULT_TTL, action=None, action_kwargs=None):
        self.key = key
        self.value = value
        self.expires = expires
        self.ttl = ttl
        self.creation_time = time.time()
        self.expire_time = self.creation_time + ttl
        self.action = action
        self.action_kwargs = action_kwargs

    def expired(self):
        return self.expires and time.time() > self.expire_time


class _KeyEncoder(json.JSONEncoder):
    def default(self, obj):
        try:
            return json.JSONEncoder.default(self, obj)
        except:
            return str(obj)


def _get_key(args, kwargs):
    return json.dumps({'args':args, 'kwargs': kwargs}, sort_keys=True, cls=_KeyEncoder)


